#!/usr/bin/env python

"""
Fix instances, removing those which are not part of the distinct set for a given
class.

Copyright (C) 2006, 2007 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

--------

To use this module, the easiest approach is to use the fix function:

fix(module)

The more complicated approach involves instantiating a Fixer object:

fixer = Fixer()

Then, applying the fixer to an existing module:

fixer.process(module)
"""

from simplify.simplified import *

# Fixing of instance information.

class Fixer(Visitor):

    """
    The name fixer which traverses the program nodes in a module, typically
    depth-first, and eliminates references to superfluous instances, replacing
    them with those from each class's distinct list, if necessary.

    See the simplify.fixnames.Fixer class for a description of the 
    """

    def __init__(self):

        "Initialise the name fixer."

        Visitor.__init__(self)

        # Satisfy visitor issues.

        self.visitor = self

    def process(self, module):

        "Process the given 'module'."

        # The fixer maintains a list of transformed subprograms (added for each
        # of the processing "roots" and also for each invoked internal
        # subprogram), along with a list of current subprograms (used to avoid
        # recursion issues) and a list of current namespaces (used to recall
        # namespaces upon invoking internal subprograms).

        self.subprograms = []
        self.current_subprograms = []

        self.module = module
        self.process_node(module)

        # Then, process all functions and methods.

        for subprogram in self.module.simplifier.subprograms:

            # Internal subprograms are skipped here and processed specially via
            # Invoke nodes.

            if not getattr(subprogram, "internal", 0):
                for specialised in subprogram.active():
                    self.subprograms.append(self.process_node(specialised))

    def process_structures(self, module):

        "Process the structures of the given 'module'."

        self.module = module

        # Visit structures and instances.

        for structure in self.module.simplifier.structures:
            for instance in structure.get_instances():
                for name, attrs in instance.namespace.items():
                    instance.namespace[name] = self._replace(attrs)

    def process_node(self, node):

        """
        Process a subprogram or module 'node', discovering from attributes on
        'node' any initial locals. Return a modified subprogram or module.
        """

        # Do not process subprograms already being processed.

        if node in self.current_subprograms:
            return None

        # Record the current subprogram.

        self.current_subprograms.append(node)

        # Dispatch to the code itself.

        result = self.dispatch(node)

        # Restore the previous subprogram and namespace.

        self.current_subprograms.pop()

        return node

    # Visitor methods.

    def default(self, node):

        """
        Process the given 'node', given that it does not have a specific
        handler.
        """

        # Process annotations.

        for name in ("non_accesses", "non_writes", "raises", "returns", "types"):
            if hasattr(node, name):
                attrs = getattr(node, name)
                setattr(node, name, self._replace(attrs))
        for name in ("accesses", "writes", "paramtypes"):
            if hasattr(node, name):
                d = getattr(node, name)
                new_d = {}
                for expr, attrs in d.items():
                    new_d[self._get_replacement(expr)] = self._replace(attrs, name)
                setattr(node, name, new_d)

        # Visit program nodes.

        for attr in ("pos_args",):
            if hasattr(node, attr):
                self.dispatches(getattr(node, attr))
        for attr in ("kw_args",):
            if hasattr(node, attr):
                self.dispatch_dict(getattr(node, attr))
        for attr in ("expr", "lvalue", "test", "star", "dstar"):
            if hasattr(node, attr):
                self.dispatch(getattr(node, attr))
        for attr in ("body", "else_", "handler", "finally_", "code", "choices", "nodes"):
            if hasattr(node, attr):
                self.dispatches(getattr(node, attr))
        if hasattr(node, "params"):
            for param, default in node.params:
                self.dispatch(default)
        for attr in ("star", "dstar"):
            if getattr(node, attr, None):
                param, default = getattr(node, attr)
                self.dispatch(default)

        return node

    def _replace(self, items, name=None):

        """
        Produce a new list or set for the given 'items', acquired from the
        annotation having the given 'name'.
        """

        if name == "accesses":
            new_items = []
        else:
            new_items = set()

        for item in list(items):
            if name == "accesses":
                attr, accessor = item
                value = attr.type
                new_items.append((Attribute(self._get_replacement(attr.context), self._get_replacement(value)), self._get_replacement(accessor)))
            else:
                attr = item
                value = attr.type
                new_items.add(Attribute(self._get_replacement(attr.context), self._get_replacement(value)))

        return new_items

    def _get_replacement(self, value):

        "Get a replacement for the given 'value'."

        # Find the distinct instance for any given instance.

        if isinstance(value, Instance):
            distinct_instances = value.get_class().get_distinct_instances()
            return distinct_instances[value]

        # For subprograms, find the distinct instance's copy for the owner
        # instance; otherwise, return the original subprogram.

        elif isinstance(value, Subprogram):
            if hasattr(value, "copy_of") and hasattr(value, "instance"):
                cls = value.instance.get_class()
                distinct = cls.get_distinct_instances()
                instance = distinct[value.instance]
                return value.copy_of.copies.get(instance, value)
            else:
                return value

        # Return all other values as they are.

        else:
            return value

    def dispatch(self, node, *args):
        return Visitor.dispatch(self, node, *args)

    def visitInvokeFunction(self, invoke):

        "Transform the 'invoke' node, performing processing on subprograms."

        return self.default(invoke)

    def visitInvokeRef(self, invoke):

        "Transform the 'invoke' node, performing processing on subprograms."

        # The special case of internal subprogram invocation is addressed by
        # propagating namespace information to the subprogram and processing it.

        subprogram = self.process_node(invoke.ref)

        if subprogram is not None:
            self.subprograms.append(subprogram)
        return invoke

# Convenience functions.

def fix_structures(module):

    "Fix the structures in the given 'module'."

    fixer = Fixer()
    fixer.process_structures(module)

def fix(module):

    "Fix the structure references in the given 'module'."

    fixer = Fixer()
    fixer.process(module)

# vim: tabstop=4 expandtab shiftwidth=4
